@@ -6,6 +6,7 @@ Hans Buehler 2021
 
 from collections import OrderedDict
 from sortedcontainers import SortedDict
+from .util import uniqueHash
 from .logger import Logger
 _log = Logger(__file__)
 
@@ -24,13 +25,13 @@ class Config(OrderedDict):
         Set data as usual:
         
             config = Config()
-            config['features']           = [ 'ime', 'spot' ]
-            config['weights']            = [ 1, 2, 3 ]
+            config['features']  = [ 'ime', 'spot' ]
+            config['weights']   = [ 1, 2, 3 ]
             
-        Create sub configuration
+        Use member notation
         
-            config.network['samples']    = 10000
-            config.network['activation'] = 'relu'
+            config.network.samples    = 10000
+            config.network.activation = 'relu'
             
     Read
         def read_config( confg ):
@@ -56,8 +57,8 @@ class Config(OrderedDict):
             config.done()                                     # no error as 'network' was detached
             network.done()                                    # error as network.activation was not read
     
-    Self-recording all available configs
-    When a value is read, one can also specify an optional help text.
+    Self-recording "help"
+    When reading a value, specify an optional help text:
     
         def read_config( confg ):
         
@@ -73,7 +74,41 @@ class Config(OrderedDict):
 
     Attributes
     ----------
-        Can access any sub config.
+        config_name : str
+            Qualified name of the config, useful for error messages
+        children : dict
+            Children of this config
+        as_dict : dict
+            Converts object into a dict of dict's'
+        recorder : dict
+            Records usage of the object and its children.
+            
+    Methods
+    -------
+        done()
+            Checks that all arguments passed to the config, and its children have been read.
+            If not, a warning will be produced. This helps catching typos.
+            The function will the call mark_done() on itself and all children
+            such that subsequent calls to done() to not produce error messages
+            
+        mark_done()
+            Marks all elements of this config and all children as 'read' (done).
+            Any elements not read explicitly will not be recorded for usage_report
+            
+        reset_done()
+            Removes all usage information for all members of this config and its children.
+            Note: this does not affected recorded previous usage.
+            
+        detach()
+            Create a copy of the current object; set it as 'done'; and return it.
+            The copy keeps the reference to the usage recorder.
+            This is used to defer using children of configs to a later stage.
+            See examples above
+            
+        copy()
+            Returns a blank copy of the current config, with a new recorder.
+    
+        
     """
         
     def __init__(self, *args, config_name = "config", **kwargs):
@@ -111,12 +146,15 @@ class Config(OrderedDict):
         return self._children
     
     def __str__(self) -> str:
-        """ Print myself via usage_report """
+        """ Print all inputs """
         return self.usage_report()
 
     @property
     def as_dict(self) -> dict:
-        """ Convert into dictionary of dictionaries """
+        """
+        Convert into dictionary of dictionaries
+        This operation will turn all members to 'read'
+        """
         d = dict(self)
         for n in self._children:
             c = self._children[n].as_dict
@@ -129,32 +167,46 @@ class Config(OrderedDict):
         """ Checks whether any variables have been set """
         return len(self) + len(self._children) == 0
     
-    def done(self, include_sub_children : bool = True ):
+    def done(self, include_children : bool = True, mark_done : bool = True ):
         """ 
-        Closes the config and checks that no unread parameters, or child configs remain.
-        If you wish to use any child config later on, use "detach()"
+        Closes the config and checks that no unread parameters remain.
+        By default this function also validates that all child configs were
+        "done".
+        If you want to make a copy of a child config for later processing
+        use detach() first
+        
+            config = Config()
+            config.a = 1
+            config.child.b = 2
+            
+            _ = config.a # read a
+            child = config.child.detach()
+            config.done()   # no error
+        
         """
         inputs = set(self)
         rest   = inputs - self._read
         if len(rest) > 0:
             _log.verify( False, "Error closing config '%s': the following config arguments were not read: %s\nRecord of this object:\n%s", \
                                         self._name, list(rest), \
-                                        self.usage_report(filter_path=self._name ) )
-        
-        if include_sub_children:
+                                        self.usage_report(filter_path=self._name ) )        
+        if include_children:
             for config in self._children:
-                self._children[config].done()
+                self._children[config].done(include_children=include_children,mark_done=False)
+        if mark_done:
+            self.mark_done(include_children=include_children)
         return
     
     def reset_done(self):
         """ Undo done """
         self._read = set()
     
-    def mark_done(self):
-        """ Announce all data was read correctly """
+    def mark_done(self, include_children : bool = True ):
+        """ Mark all members as being read. Once called calling done() will no longer trigger an error """
         self._read = set( self )
-        for c in self._childen:
-            c.mark_done()
+        if include_children: 
+            for c in self._childen:
+                c.mark_done(include_children=include_children)
 
     def __getattr__(self, key : str):
         """
@@ -176,7 +228,7 @@ class Config(OrderedDict):
         as "done" unless 'mark_self_done' is used.
         The copy will have the same shared recorder as 'self' unless new_recorder is True.
         
-        The use case for this is storing sub config's for later processing
+        The use case for this function is storing sub config's for later processing
         Example:
             
             Creates
@@ -241,14 +293,17 @@ class Config(OrderedDict):
             
         Parameters
         ----------
-            other : Config, optional
+            other : dict, Config, optional
                 Copy all content of 'other' into 'self'.
-                Will set all recorders to the top level element accordingly.
+                If 'other' is a config: elements of other will /not/ be marked as used, or recorded.
+                Use mark_used() if this is desired
             **kwargs
                 Allows assigning specific values.
         """
         if not other is None:
             if isinstance( other, Config ):
+                # copy() children
+                # and reset recorder to ours.
                 def set_recorder(config, recorder):
                     config._recorder = recorder
                     for c in config._children:
@@ -258,6 +313,8 @@ class Config(OrderedDict):
                     child = child.copy()
                     set_recorder(child, self._recorder)
                     self._children[sub]= child
+            # copy elements from other.
+            # we do not mark elements from another config as 'used' 
             for key in other:
                 self[key] = OrderedDict.get(other,key)
         OrderedDict.update( self, kwargs )
@@ -271,7 +328,8 @@ class Config(OrderedDict):
                        help         : str = None, 
                        help_default : str = None, 
                        help_cast    : str = None, 
-                       mark_read    : bool = True ):
+                       mark_read    : bool = True,
+                       record       : bool = True ):
         """
         Reads 'key' from the config. If not found, return 'default' if specified.
     
@@ -298,6 +356,8 @@ class Config(OrderedDict):
                 Use this for complex cast types which are hard to read.
             mark_read : bool, optional
                 If true, marks the respective element as read.
+            record : bool, optional
+                If true, records usage of the key. 
                 
         Returns
         -------
@@ -319,7 +379,11 @@ class Config(OrderedDict):
         if mark_read:
             self._read.add(key)
         
-        # record
+        # avoid recording
+        if not record:
+            return value
+
+        # record?
         record_key    = self._name + "['" + key + "']"
         help          = str(help) if not help is None and len(help) > 0 else ""
         help          = help[:-1] if help[-1:] == "." else help
@@ -346,7 +410,7 @@ class Config(OrderedDict):
                     _log.verify( exst_value == record, "Config %s was used twice with different default/help values. Found %s and %s, respectively", record_key, exst_value, record )
         else:
             self._recorder[record_key] = record
-    
+
         # done
         return value
         
@@ -354,19 +418,23 @@ class Config(OrderedDict):
         """ Returns self(key)  """
         return self(key)
     def get(self, key : str, default = None ):
-        """ Returns self(key) """
+        """ Returns self(key, default) """
         return self(key, default)
     def get_default(self, key : str, default):
         """ Returns self(key,default) """
         return self(key,default)
+    def get_raw(self, key : str, default = None ):
+        """ Returns self(key, default, mark_read=False, record=False ) """
+        return self(key, default, mark_read=False, record=False)
     
     # Write
     # -----
     
     def __setattr__(self, key, value):
-        """ Assign value like self[key] = value
-            If 'value' is a config, then the config will be assigned as child.
-            Its recorder and name will be overwritten.
+        """
+        Assign value like self[key] = value
+        If 'value' is a config, then the config will be assigned as child.
+        Its recorder and name will be overwritten.
         """
         if key[0] == "_" or key in self.__dict__:
             OrderedDict.__setattr__(self, key, value )
@@ -388,10 +456,10 @@ class Config(OrderedDict):
     def usage_report(self,    with_values  : bool = True,
                               with_help    : bool = True,
                               with_defaults: bool = True,
-                              with_cast   : bool = False,
+                              with_cast    : bool = False,
                               filter_path  : str  = None ) -> str:
         """
-        Generate a human readable config report of the top leverl recorder
+        Generate a human readable report of all variables read from this config.
     
         Parameters
         ----------
@@ -420,7 +488,7 @@ class Config(OrderedDict):
         with_values   = bool(with_values)
         with_help     = bool(with_help)
         with_defaults = bool(with_defaults)
-        with_cast    = bool(with_cast)
+        with_cast     = bool(with_cast)
         l             = len(filter_path) if not filter_path is None else 0
         rep_here      = ""
         reported      = ""
@@ -457,7 +525,7 @@ class Config(OrderedDict):
 
     def usage_reproducer(self):
         """
-        Prints an expression which will reproduce the current
+        Returns an expression which will reproduce the current
         configuration tree as long as each 'value' handles
         repr() correctly.
         """
@@ -468,6 +536,45 @@ class Config(OrderedDict):
             report       += key + " = " + repr(value) + "\n"
         return report
         
+    def input_report(self):
+        """
+        Returns a report of all inputs in a readable format, as long as all values
+        are as such.
+        """
+        inputs = []      
+        def ireport(self, inputs):
+            for key in self:
+                value      = self.get_raw(key)
+                report_key = self._name + "['" + key + "'] = %s" % str(value)
+                inputs.append( report_key )
+            for c in self._children:
+                ireport(self._children[c], inputs)
+        ireport(self, inputs)
+        
+        inputs = sorted(inputs)
+        report = ""
+        for i in inputs:
+            report += i + "\n"        
+        return report
+        
+    def unique_id(self):
+        """ Returns an MDH5 hash key for this object, based on 'input_report()'
+            ** WARNING **
+            This function ignores any config keys or children with leading '_'.
+        """
+        inputs = []      
+        for key in self:
+            if key[:1] == "_":
+                continue
+            value      = self.get_raw(key)
+            hash_      = uniqueHash(key,uniqueHash(value))
+            inputs.append( hash_ )
+        for c in self._children:
+            if c[:1] == "_":
+                continue
+            inputs.append( uniqueHash(c,self._children[c].unique_id) )
+        return uniqueHash(inputs)
+
     # magic
     # -----
     
